import json
import requests
from typing import Union

import pprint
import pandas as pd

from ..._common.config import *
from ..._common.const import *
from prism import _core
from ..._utils import (
    _validate_args,
    get,
    post,
    patch,
    delete,
    _authentication,
    _process_fileresponse,
    _fetch_and_parse,
    update_query,
    Loader,
    download,
    plot_tree,
)
from ..._prismcomponent import prismcomponent as pcmpt
from ._universe import parse_universe_to_universeid


__all__ = [
    'list_dataquery',
    'load_dataquery',
    'save_dataquery',
    'rename_dataquery',
    'delete_dataquery',
    'update_dataquery',
    'get_data',
]


@_validate_args
def list_dataquery(search: str = None, tree=False):
    dataqueries_df = _fetch_and_parse(URL_DATAQUERIES, 'pdq', search)
    if dataqueries_df.empty:
        raise SystemExit('No Dataqueries found.')
    dataqueries_df = dataqueries_df[dataqueries_df['folderflag']==0]
    if not tree:
        return dataqueries_df.drop('folderflag', axis=1)
    plot_tree(dataqueries_df['dataqueryname'].tolist(), 'dataquery')


def load_dataquery(dataquery: Union[str, int]):
    headers = _authentication()
    dataqueryid = parse_dataquery_to_dataqueryid(dataquery)
    res = requests.get(url=URL_DATAQUERIES + f'/{dataqueryid}', headers=headers)
    if not res.ok:
        raise SystemExit(res.json()['message'])
    query = res.json()['rescontent']['data'][0]['dataquerybody']
    return _dataquery_to_component(query)


def _dataquery_to_component(query: dict):
    component_name = query['component_name']
    category_component_info = COMPONENT2CATEGORY.get(component_name)
    if category_component_info is None:
        component = pcmpt._PrismComponent(**query)
    else:
        category_name = category_component_info['category']
        component_fn_name = category_component_info['component']
        category = getattr(_core, category_name)
        component_fn = getattr(category, component_fn_name)
        component = component_fn(**query['component_args'])
    return component


def rename_dataquery(old: Union[str, int], new: str):
    should_overwrite, err_msg = should_overwrite_dataquery(new)
    if not should_overwrite:
        raise SystemExit(err_msg)
    dataqueryid = parse_dataquery_to_dataqueryid(old)
    ret = patch(URL_DATAQUERIES + f'/{dataqueryid}', {'newpath': new + '.pdq'}, None)
    print(ret['message'])
    return ret

@_validate_args
def save_dataquery(component, dataqueryname: str):
    should_overwrite, err_msg = should_overwrite_dataquery(dataqueryname)
    if not should_overwrite:
        print(f'{err_msg}')
        return

    ret = post(URL_DATAQUERIES, {'path': dataqueryname + '.pdq'}, component._query)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


def delete_dataquery(dataquery: Union[str, int]):
    dataqueryid = parse_dataquery_to_dataqueryid(dataquery)
    ret = delete(URL_DATAQUERIES + f'/{dataqueryid}')
    print(f'{ret["message"]}')
    return ret


@_validate_args
def datacomponent_args_info(component_name: str):
    return get(URL_COMPONENT + f'/datacomponent/{component_name}/args')


def update_dataquery(component):
    print(update_query(component))
    # return update_query(component)


@_validate_args
def get_data(
    obj,
    universe=None,
    startdate: str = None,
    enddate: str = None,
    shownid: list = None,
) -> pd.DataFrame:
    query = []
    cmpts = set()
    if not isinstance(obj, list):
        obj = [obj]
    for o in obj:
        if isinstance(o, pcmpt._PrismComponent):
            query.append(o._query)
            cmpts.add(o._query['component_name'])
        else:
            raise ValueError('Please provide Components into get data.', True, None)
    if len(cmpts - set(UniverseFreeDataComponentType)) == 0:
        universeid = None
    else:
        universeid = parse_universe_to_universeid(universe)
    default_shownid = True
    if (shownid is not None) and (len(shownid) == 0):
        shownid = None
        default_shownid = False
    params = {
        'universeid': universeid,
        'startdate': startdate,
        'enddate': enddate,
        'shownid': shownid,
        'default_shownid': default_shownid,
    }
    headers = _authentication()

    with Loader("\rFetching Data...") as l:
        res = requests.post(url=URL_DATAQUERIES + '/query', params=params, json=query, headers=headers, stream=True)
        if res.status_code >= 400:
            return _process_fileresponse(res, 'data')
        rescontent = download(res)

    return _process_fileresponse(res, 'data', rescontent)


def parse_dataquery_to_dataqueryid(dataquery: Union[str, int]):
    dataqueries_df = _fetch_and_parse(URL_DATAQUERIES, 'pdq')
    if dataqueries_df.empty:
        raise SystemExit('No dataqueries Found.')

    if isinstance(dataquery, int):
        dataqueries_df = dataqueries_df[dataqueries_df['dataqueryid'] == dataquery]
    elif isinstance(dataquery, str):
        dataqueries_df = dataqueries_df[dataqueries_df['dataqueryname'] == dataquery]
    else:
        raise SystemExit('Please provide dataquery name or dataquery id for dataquery.')

    if dataqueries_df.empty:
        raise SystemExit(f'No dataquery matching: {dataquery}')

    dataqueryid = dataqueries_df['dataqueryid'].values[0]
    return dataqueryid


def should_overwrite_dataquery(dataqueryname):
    dataqueries_df = _fetch_and_parse(URL_DATAQUERIES, 'pdq')
    if dataqueries_df.empty:
        return True, None

    if dataqueryname in dataqueries_df['dataqueryname'].to_list():
        overwrite = input(f'{dataqueryname} already exists. Do you want to overwrite? (Y/N) \n')
        if overwrite.lower() == 'y':
            return True, None
        elif overwrite.lower() == 'n':
            return False, 'Not saving dataquery.'
        else:
            return False, 'Dataquery not saved. Please provide a valid input.'
    else:
        return True, None
