import copy
from xml.etree.ElementTree import TreeBuilder
import requests
from typing import Union

import prism
from ..._common.config import *
from ..._common.const import *
from ..._core._req_builder import _dataquery
from ..._utils import _validate_args, post, patch, delete, _authentication, _fetch_and_parse, plot_tree
from ..._prismcomponent import prismcomponent as pcmpt


__all__ = ['list_taskquery', 'load_taskquery', 'save_taskquery', 'delete_taskquery', 'rename_taskquery']


@_validate_args
def list_taskquery(search: str = None, tree=False):
    taskqueries_df = _fetch_and_parse(URL_TASKQUERIES, 'ptq', search)
    if taskqueries_df.empty:
        raise SystemExit('No Taskqueries found.')
    if not tree:
        return taskqueries_df.drop('folderflag', axis=1)
    plot_tree(taskqueries_df['taskqueryname'].tolist(), 'taskquery')



def load_taskquery(taskquery: Union[str, int]):
    headers = _authentication()
    taskqueryid = parse_taskquery_to_taskqueryid(taskquery)
    res = requests.get(url=URL_TASKQUERIES + f'/{taskqueryid}', headers=headers)
    if not res.ok:
        raise SystemExit(res.json()['message'])
    query = res.json()['rescontent']['data'][0]['taskquerybody']
    component_name = query.pop('component_name')
    component = getattr(prism, component_name)
    component_args = query['component_args']
    for k in copy.deepcopy(list(component_args.keys())):
        if '_dataquery' in k:
            arg_key = k.split('_dataquery')[0]
            dataquery = _dataquery._dataquery_to_component(component_args.pop(k))
            component_args[arg_key] = dataquery
    result = component(**component_args)
    return result


@_validate_args
def save_taskquery(component, taskqueryname: str):
    should_overwrite, err_msg = should_overwrite_taskquery(taskqueryname)
    if not should_overwrite:
        raise SystemExit(err_msg)
    ret = post(URL_TASKQUERIES, {'path': taskqueryname + '.ptq',}, component._query)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


def rename_taskquery(old: Union[str, int], new: str):
    should_overwrite, err_msg = should_overwrite_taskquery(new)
    if not should_overwrite:
        raise SystemExit(err_msg)
    taskqueryid = parse_taskquery_to_taskqueryid(old)
    ret = patch(URL_TASKQUERIES + f'/{taskqueryid}', {'newpath': new + 'ptq'}, None)
    print(ret['message'])
    return ret


def delete_taskquery(dataquery: Union[str, int]):
    dataqueryid = parse_taskquery_to_taskqueryid(dataquery)
    ret = delete(URL_TASKQUERIES + f'/{dataqueryid}')
    print(ret['message'])
    return ret


def parse_taskquery_to_taskqueryid(taskquery: Union[str, int]):
    taskqueries_df = _fetch_and_parse(URL_TASKQUERIES, 'ptq')
    if taskqueries_df.empty:
        raise SystemExit('No taskquery found.')

    if isinstance(taskquery, int):
        taskqueries_df = taskqueries_df[taskqueries_df['taskqueryid'] == taskquery]
    elif isinstance(taskquery, str):
        taskqueries_df = taskqueries_df[taskqueries_df['taskqueryname'] == taskquery]
    else:
        raise SystemExit('Please provide taskquery path or taskquery id for taskquery.')

    if taskqueries_df.empty:
        raise SystemExit(f'No taskquery matching: {taskquery}')

    taskqueryid = taskqueries_df['taskqueryid'].values[0]
    return taskqueryid


def should_overwrite_taskquery(taskqueryname):
    taskqueries_df = _fetch_and_parse(URL_TASKQUERIES, 'ptq')
    if taskqueries_df.empty:
        return True, None

    if taskqueryname in taskqueries_df['taskqueryname'].to_list():
        overwrite = input(f'{taskqueryname} already exists. Do you want to overwrite? (Y/N) \n')
        if overwrite.lower() == 'y':
            return True, None
        elif overwrite.lower() == 'n':
            return False, 'Not saving taskquery.'
        else:
            return False, 'Please provide a valid input.'
    else:
        return True, None