import requests
from typing import Union

from ..._common.config import *
from ..._common.const import *
from ._list import _list_dataitem_index
from ..._utils import (
    _validate_args,
    post,
    patch,
    delete,
    _authentication,
    _process_response,
    _process_fileresponse,
    _fetch_and_parse,
    plot_tree,
)


__all__ = [
    'list_universe',
    'get_universe',
    'upload_universe',
    'save_index_as_universe',
    'combine_universe',
    'rename_universe',
    'delete_universe',
    'get_universe_template',
]


@_validate_args
def list_universe(search: str = None, tree=False):
    universes_df = _fetch_and_parse(URL_UNIVERSES, 'puv', search)
    if universes_df.empty:
        raise SystemExit('No Universes found.')
    if not tree:
        return universes_df.drop('folderflag', axis=1)
    plot_tree(universes_df['universename'].tolist(), 'universe')


@_validate_args
def get_universe(universe, startdate: str = None, enddate: str = None, shownid: list = None, expand : bool = False):
    universeid = parse_universe_to_universeid(universe)
    params = {'startdate': startdate, 'enddate': enddate, 'shownid': shownid, 'expand': expand}
    headers = _authentication()
    res = requests.get(URL_UNIVERSES + f'/{universeid}', params=params, headers=headers)
    return _process_fileresponse(res, 'universe')


@_validate_args
def upload_universe(universefile: str, universename: str):
    universefile_params = {'universefile': open(universefile, 'rb')}
    headers = {'Authorization': _authentication()['Authorization'], 'Client-Channel': 'python-extension'}
    res = requests.post(
        url=URL_UNIVERSES + "/flatfile",
        params={'path': universename + '.puv'},
        files=universefile_params,
        headers=headers
    )
    ret = _process_response(res)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def save_index_as_universe(
    dataitemid: int,
    startdate: str = None,
    enddate: str = None,
    universename: str = None,
):
    if universename is None:
        indices = _list_dataitem_index(datacomponent='Index Level')
        universename = indices[indices['dataitemid'] == dataitemid]['dataitemname'].values[0]

    should_overwrite, err_msg = should_overwrite_universe(universename)
    if not should_overwrite:
        raise SystemExit(err_msg)
    params = {
        'dataitemid': dataitemid,
        'startdate': startdate,
        'enddate': enddate,
        'path': universename + '.puv',
    }
    ret = post(URL_UNIVERSES + f'/index', params, None)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def combine_universe(universes: list, newuniversename: str):
    if len(universes) < 2:
        SystemExit('Please provide more than 2 universes to combine.')
    should_overwrite, err_msg = should_overwrite_universe(newuniversename)
    if not should_overwrite:
        raise SystemExit(err_msg)
    universeids = parse_universes_to_universeids(universes)
    params = {"universeids": universeids, "path": newuniversename + '.puv'}
    ret = post(URL_UNIVERSES + f'/combine', params, None)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


def rename_universe(old: Union[str, int], new: str):
    should_overwrite, err_msg = should_overwrite_universe(new)
    if not should_overwrite:
        SystemExit(err_msg)
    universeid = parse_universe_to_universeid(old)
    ret = patch(URL_UNIVERSES + f'/{universeid}', {'newpath': new + '.puv'}, None)
    print(ret['message'])
    return ret


def delete_universe(universe: Union[str, int]):
    universeid = parse_universe_to_universeid(universe)
    ret = delete(URL_UNIVERSES + f'/{universeid}')
    print(ret['message'])
    return ret


def get_universe_template():
    res = requests.get(url=URL_UNIVERSES + '/template', headers=_authentication())
    return res.text


def parse_universe_to_universeid(universe: Union[str, int]):
    universes_df = _fetch_and_parse(URL_UNIVERSES, 'puv')
    if universes_df.empty:
        raise ValueError('No universe found.', True, None)

    if isinstance(universe, int):
        universes_df = universes_df[universes_df['universeid'] == universe]
    elif isinstance(universe, str):
        universes_df = universes_df[universes_df['universename'] == universe]
    else:
        raise ValueError('Please provide universe name or universe id for universe.', True, None)

    if universes_df.empty:
        raise ValueError(f'No universe matching: {universe}', True, None)
    universeid = universes_df['universeid'].values[0]

    return universeid


def parse_universes_to_universeids(universes: list):
    universes_df = _fetch_and_parse(URL_UNIVERSES, 'puv')
    if universes_df.empty:
        raise ValueError('No universe found.', True, None)

    universenames = []
    universeids = []
    for u in universes:
        if isinstance(u, int):
            universeids.append(u)
        elif isinstance(u, str):
            universenames.append(u)
        else:
            raise ValueError('Please provide universe name or universe id for universe.', True, None)

    ret_universeids = set()
    if len(universenames) != 0:
        matching_universes_df = universes_df[universes_df['universename'].isin(universenames)]
        if len(matching_universes_df) != len(universenames):
            not_matched = list(set(universenames) - set(matching_universes_df['universename'].tolist()))
            raise ValueError(f'No universe matching: {not_matched}', True, None)
        universeids_from_universenames = matching_universes_df['universeid'].tolist()
        ret_universeids.update(universeids_from_universenames)

    if len(universeids) != 0:
        matching_universes_df = universes_df[universes_df['universeid'].isin(universeids)]
        if len(matching_universes_df) != len(universeids):
            not_matched = list(set(universeids) - set(matching_universes_df['universeid'].tolist()))
            raise ValueError(f'No universe matching: {not_matched}', True, None)
        existing_universeids = matching_universes_df['universeid'].tolist()
        ret_universeids.update(existing_universeids)
    return list(ret_universeids)


def should_overwrite_universe(universename):
    universes_df = _fetch_and_parse(URL_UNIVERSES, 'puv')
    if universes_df.empty:
        return True, None

    if universename in universes_df['universename'].to_list():
        overwrite = input(f'{universename} already exists. Do you want to overwrite? (Y/N) \n')
        if overwrite.lower() == 'y':
            return True, None
        elif overwrite.lower() == 'n':
            return False, 'Not saving universe.'
        else:
            return False, 'Please provide a valid input.'
    else:
        return True, None
